###############################################################################
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
###############################################################################

# executable test suite as per WTH, Annex A

import json
import logging
from uuid import UUID

import click

from pywis_topics.centre_id import CentreId
from pywis_topics.cli_options import cli_common_options
from pywis_topics.topics import TopicHierarchy
from pywis_topics.util import get_current_datetime_rfc3339


LOGGER = logging.getLogger(__name__)


def gen_test_id(test_id: str) -> str:
    """
    Convenience function to print test identifier as URI

    :param test_id: test suite identifier

    :returns: test identifier as URI
    """

    return f'http://wis.wmo.int/spec/wth/a/conf/core/{test_id}'


class WTHTestSuite:
    """Test suite for WIS2 Topic Hierarchy"""

    def __init__(self, data: dict):
        """
        initializer

        :param data: str of topic hierarchy

        :returns: `pywis_topics.ets.WTHTestSuite`
        """

        self.test_id = None
        self.topic = data
        self.report = []

    def run_tests(self):
        """Convenience function to run all tests"""

        results = []
        tests = []
        ets_report = {
            'summary': {},
        }

        for f in dir(WTHTestSuite):
            if all([
                    callable(getattr(WTHTestSuite, f)),
                    f.startswith('test_requirement')]):

                tests.append(f)

        for t in tests:
            results.append(getattr(self, t)())

        for code in ['PASSED', 'FAILED', 'SKIPPED']:
            r = len([t for t in results if t['code'] == code])
            ets_report['summary'][code] = r

        ets_report['tests'] = results
        ets_report['datetime'] = get_current_datetime_rfc3339()

        return {
            'ets-report': ets_report
        }

    def test_requirement_id(self):
        """
        Check for the existence of a valid id property.
        """

        status = {
            'id': gen_test_id('id'),
            'code': 'PASSED',
        }

        try:
            UUID(self.message['id'])
        except ValueError as err:
            status['code'] = 'FAILED'
            status['message'] = f'Invalid UUID: {err}'

        return status


@click.group()
def ets():
    """executable test suite"""
    pass


@click.command()
@click.pass_context
@get_cli_common_options
@click.argument('topic')
def validate(ctx, topic, logfile, verbosity):
    """validate against the abstract test suite"""

    click.echo(f'Testing {topic}')

    ts = WTHTestSuite(topic)
    try:
        results = ts.run_tests()
    except Exception as err:
        raise click.ClickException(err)

    click.echo(json.dumps(results, indent=4))
    ctx.exit(results['ets-report']['summary']['FAILED'])


ets.add_command(validate)



###############################################################################
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
###############################################################################

import csv
import logging
from pathlib import Path

import click

from pywis_topics.topics import TopicHierarchy, validate_baseline
from pywis_topics.util import get_cli_common_options, get_userdir, setup_logger

LOGGER = logging.getLogger(__name__)


WIS2_TOPIC_HIERARCHY_LOOKUP = Path(get_userdir()) / 'wis2-topic-hierarchy'


class CentreId:
    def __init__(self, centre_id: str, tables: str = None):
        """
        Initializer

        :param centre_id: centre-id as defined by WTH
        :param tables: location of base directory for bundle

        :returns: `pywis_topics.centre_id.CentreId`
        """

        self.centre_id = centre_id
        try:
            self.tld, self.centre = self.centre_id.split('-', 1)
        except ValueError:
            msg = 'Invalid number of centre-id tokens'
            raise ValueError(msg)

        if tables is not None:
            self.tables_dir = Path(tables) / 'wis2-topic-hierarchy'
        else:
            self.tables_dir = WIS2_TOPIC_HIERARCHY_LOOKUP

    def validate(self) -> bool:
        """
        Validates a centre-id

        :returns: `bool` of whether topic hierarchy is valid
        """

        LOGGER.debug('Validating baseline conventions')
        if not validate_baseline(self.centre_id):
            return False

        LOGGER.debug('Validating TLD component')
        tld_valid = False
        tld_upper = self.tld.upper()

        tld_file = self.tables_dir / 'tlds-alpha-by-domain.txt'
        with tld_file.open() as fh:
            reader = csv.reader(fh)
            next(reader)
            for row in reader:
                if tld_upper == row[0]:
                    tld_valid = True
                    break

        if not tld_valid:
            LOGGER.warning('Invalid TLD')
            return False

        LOGGER.debug('Checking for uniqueness')
        topics = TopicHierarchy()
        if self.centre_id in topics.topics[3]:
            LOGGER.warning('centre-id is already allocated')
            return False

        return True


@click.group('centre-id')
def centre_id():
    """Centre id utilities"""
    pass


@click.command()
@click.pass_context
@get_cli_common_options
@click.argument('centre-id')
def validate(ctx, centre_id, logfile, verbosity):
    """Validate centre-id"""

    setup_logger(verbosity, logfile)

    try:
        cid = CentreId(centre_id)
    except ValueError as err:
        raise click.ClickException(err)

    if cid.validate():
        click.echo('Valid')
    else:
        click.echo('Invalid')


centre_id.add_command(validate)



###############################################################################
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
###############################################################################

# pywis-topics as a service
# -------------------------
#
# This file is intended to be used as a pygeoapi process plugin which will
# provide pywis-topics functionality via OGC API - Processes.
#
# To integrate this plugin in pygeoapi:
#
# 1. ensure pywis-topics is installed into the pygeoapi deployment environment
#
# 2. add the processes to the pygeoapi configuration as follows:
#
# pywis-topics-list:
#     type: process
#     processor:
#         name: pywis_topics.pygeoapi_plugin.WIS2TopicHierarchyListTopicsProcessor  # noqa
#
# pywis-topics-validate:
#     type: process
#     processor:
#         name: pywis_topics.pygeoapi_plugin.WIS2TopicHierarchyListTopicsProcessor  # noqa
#
# 3. (re)start pygeoapi
#
# The resulting processes will be available at the following endpoints:
#
# /processes/pywis-topics-wis2-topics-list
#
# /processes/pywis-topics-wis2-topics-validate
#
# Note that pygeoapi's OpenAPI/Swagger interface (at /openapi) will also
# provide a developer-friendly interface to test and run requests
#


import logging

from pygeoapi.process.base import BaseProcessor, ProcessorExecuteError

from pywis_topics.topics import TopicHierarchy

LOGGER = logging.getLogger(__name__)

WIS2_TOPIC_HIERARCHY_LINKS = [{
    'type': 'text/html',
    'rel': 'about',
    'title': 'information',
    'href': 'https://github.com/wmo-im/wis2-topic-hierarchy',
    'hreflang': 'en-US'
}]

WIS2_TOPIC_HIERARCHY_INPUT_TOPIC = {
    'title': 'Topic',
    'description': 'Topic (full or partial)',
    'schema': {
        'type': 'string'
    },
    'minOccurs': 1,
    'maxOccurs': 1,
    'metadata': None,
    'keywords': ['topic']
}


PROCESS_LIST_TOPICS = {
    'version': '0.1.0',
    'id': 'pywis-topics-list',
    'title': {
        'en': 'List WIS2 topics'
    },
    'description': {
        'en': 'Lists WIS2 topics'
    },
    'keywords': ['wis2', 'topics', 'metadata'],
    'links': WIS2_TOPIC_HIERARCHY_LINKS,
    'inputs': {
        'topic': WIS2_TOPIC_HIERARCHY_INPUT_TOPIC,
    },
    'outputs': {
        'result': {
            'title': 'List of child topics',
            'description': 'List of child topics',
            'schema': {
                'type': 'object',
                'contentMediaType': 'application/json',
                'properties': {
                    'topics': {
                        'type': 'array',
                        'minOccurs': 1,
                        'items': {
                            'type': 'string',
                            'descriptiopn': 'Matching topic'
                        }
                    }
                }
            }
        }
    },
    'example': {
        'inputs': {
            'topic': 'origin/a/wis2'
        }
    }
}


PROCESS_VALIDATE_TOPIC = {
    'version': '0.1.0',
    'id': 'pywis-topics-validate',
    'title': {
        'en': 'Validate WIS2 topics'
    },
    'description': {
        'en': 'Validates WIS2 topics'
    },
    'keywords': ['wis2', 'topics', 'metadata'],
    'links': WIS2_TOPIC_HIERARCHY_LINKS,
    'inputs': {
        'topic': WIS2_TOPIC_HIERARCHY_INPUT_TOPIC,
    },
    'outputs': {
        'result': {
            'title': 'Result of topic validity',
            'description': 'Result of topic validity',
            'schema': {
                'type': 'object',
                'contentMediaType': 'application/json',
                'properties': {
                    'topic_is_valid': {
                        'type': 'boolean'
                    }
                },
                'required': ['topic_is_valie']
            }
        }
    },
    'example': {
        'inputs': {
            'topic': 'origin/a/wis2'
        }
    }
}


class WIS2TopicHierarchyListTopicsProcessor(BaseProcessor):
    """WIS2 topic hierarchy list process"""

    def __init__(self, processor_def):
        """
        Initialize object

        :param processor_def: provider definition

        :returns: pywis_topics.pygeoapi_plugin.WIS2TopicHierarchyListTopicsProcessor  # noqa
        """

        super().__init__(processor_def, PROCESS_LIST_TOPICS)

    def execute(self, data):

        response = None
        mimetype = 'application/json'
        topic = data.get('topic')

        if topic is None:
            msg = 'Missing topic'
            LOGGER.error(msg)
            raise ProcessorExecuteError(msg)

        LOGGER.debug('Querying topic')
        th = TopicHierarchy()
        try:
            matching_topics = th.list_children(topic)
            response = {
                'topics': matching_topics
            }
        except ValueError as err:
            raise ProcessorExecuteError(err)

        return mimetype, response

    def __repr__(self):
        return '<WIS2TopicHierarchyListTopicsProcessor>'


class WIS2TopicHierarchyValidateTopicProcessor(BaseProcessor):
    """WIS2 topic hierarchy validate process"""

    def __init__(self, processor_def):
        """
        Initialize object

        :param processor_def: provider definition

        :returns: pywis_topics.pygeoapi_plugin.WIS2TopicHierarchyValidateTopicProcessor  # noqa
        """

        super().__init__(processor_def, PROCESS_VALIDATE_TOPIC)

    def execute(self, data):

        response = None
        mimetype = 'application/json'
        topic = data.get('topic')

        if topic is None:
            msg = 'Missing topic'
            LOGGER.error(msg)
            raise ProcessorExecuteError(msg)

        LOGGER.debug('Querying topic')
        th = TopicHierarchy()
        response = {
            'topic_is_valid': th.validate(topic)
        }
        return mimetype, response

    def __repr__(self):
        return '<WIS2TopicHierarchyValidateTopicProcessor>'



###############################################################################
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
###############################################################################

import io
import logging
import os
import shutil
import zipfile

import click

from pywis_topics.util import (get_cli_common_options, get_userdir, urlopen_,
                               setup_logger)

LOGGER = logging.getLogger(__name__)

USERDIR = get_userdir()

WIS2_TOPIC_HIERARCHY_DIR = get_userdir() / 'wis2-topic-hierarchy'


@click.group()
def bundle():
    """Configuration bundle management"""
    pass


def sync_bundle() -> None:
    """
    Sync bundle locally to ~/.pywis-topics

    :returns: `None`
    """

    LOGGER.debug('Caching topic hierarchy')

    if USERDIR.exists():
        shutil.rmtree(USERDIR)

    LOGGER.debug('Downloading WIS2 topic hierarchy')
    WIS2_TOPIC_HIERARCHY_DIR.mkdir(parents=True, exist_ok=True)

    ZIPFILE_URL = 'https://wmo-im.github.io/wis2-topic-hierarchy/wth-bundle.zip'  # noqa
    FH = io.BytesIO(urlopen_(ZIPFILE_URL).read())
    with zipfile.ZipFile(FH) as z:
        LOGGER.debug(f'Processing zipfile "{z.filename}"')
        for name in z.namelist():
            LOGGER.debug(f'Processing entry "{name}"')
            filename = os.path.basename(name)

            dest_file = WIS2_TOPIC_HIERARCHY_DIR / filename
            LOGGER.debug(f'Creating "{dest_file}"')
            with z.open(name) as src, dest_file.open('wb') as dest:
                shutil.copyfileobj(src, dest)

    LOGGER.debug('Downloading IANA TLDs')
    IANA_URL = 'https://data.iana.org/TLD/tlds-alpha-by-domain.txt'
    iana_file = WIS2_TOPIC_HIERARCHY_DIR / 'tlds-alpha-by-domain.txt'
    with iana_file.open('wb') as fh:
        fh.write(urlopen_(f'{IANA_URL}').read())


@click.command()
@get_cli_common_options
@click.pass_context
def sync(ctx, logfile, verbosity):
    "Sync configuration bundle"""

    setup_logger(verbosity, logfile)
    sync_bundle()


bundle.add_command(sync)



###############################################################################
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
###############################################################################

import logging
import sys

import click


def OPTION_VERBOSITY(f):
    logging_options = ['ERROR', 'WARNING', 'INFO', 'DEBUG']

    def callback(ctx, param, value):
        if value is not None:
            logging.basicConfig(stream=sys.stdout,
                                level=getattr(logging, value))
        return True

    return click.option('--verbosity', '-v',
                        type=click.Choice(logging_options),
                        help='Verbosity',
                        callback=callback)(f)


def cli_callbacks(f):
    f = OPTION_VERBOSITY(f)
    return f



###############################################################################
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
###############################################################################

from pathlib import Path
import re
from setuptools import Command, find_packages, setup
import sys


class PyTest(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import subprocess
        errno = subprocess.call([sys.executable, 'tests/run_tests.py'])
        raise SystemExit(errno)


def read(filename, encoding='utf-8'):
    """read file contents"""

    fullpath = Path(__file__).resolve().parent / filename

    with fullpath.open() as fh:
        contents = fh.read().strip()
    return contents


def get_package_version():
    """get version from top-level package init"""
    version_file = read('pywis_topics/__init__.py')
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


LONG_DESCRIPTION = read('README.md')

DESCRIPTION = 'pywis-topics is a utility to work with the WIS2 Topic Hierarchy'

MANIFEST = Path('MANIFEST')

if MANIFEST.exists():
    MANIFEST.unlink()


setup(
    name='pywis-topics',
    version=get_package_version(),
    description=DESCRIPTION.strip(),
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    license='Apache Software License',
    platforms='all',
    keywords=' '.join([
        'WIS2',
        'WTH',
        'topics'
    ]),
    author='Tom Kralidis',
    author_email='tomkralidis@gmail.com',
    maintainer='Tom Kralidis',
    maintainer_email='tomkraldis@gmail.com',
    url='https://github.com/wmo-im/pywis-topics',
    install_requires=read('requirements.txt').splitlines(),
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'pywis-topics=pywis_topics:cli'
        ]
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python'
    ],
    cmdclass={'test': PyTest}
)



###############################################################################
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
###############################################################################

import logging
from pathlib import Path
import ssl
import sys
from typing import Union

from urllib.error import URLError
from urllib.request import urlopen

LOGGER = logging.getLogger(__name__)


def get_typed_value(value) -> Union[float, int, str]:
    """
    Derive true type from data value

    :param value: value

    :returns: value as a native Python data type
    """

    try:
        if '.' in value:  # float?
            value2 = float(value)
        elif len(value) > 1 and value.startswith('0'):
            value2 = value
        else:  # int?
            value2 = int(value)
    except ValueError:  # string (default)?
        value2 = value

    return value2


def get_userdir() -> str:
    """
    Helper function to get userdir

    :returns: user's home directory
    """

    return Path.home() / '.pywis-topics'


def get_cli_common_options(function):
    """
    Define common CLI options
    """

    import click
    function = click.option('--verbosity', '-v',
                            type=click.Choice(
                                ['ERROR', 'WARNING', 'INFO', 'DEBUG']),
                            help='Verbosity')(function)
    function = click.option('--log', '-l', 'logfile',
                            type=click.Path(writable=True, dir_okay=False),
                            help='Log file')(function)
    return function


def setup_logger(loglevel: str = None, logfile: str = None) -> None:
    """
    Setup logging

    :param loglevel: logging level
    :param logfile: logfile location

    :returns: void (creates logging instance)
    """

    if loglevel is None and logfile is None:  # no logging
        return

    if loglevel is None and logfile is not None:
        loglevel = 'INFO'

    log_format = \
        '[%(asctime)s] %(levelname)s - %(message)s'
    date_format = '%Y-%m-%dT%H:%M:%SZ'

    loglevels = {
        'CRITICAL': logging.CRITICAL,
        'ERROR': logging.ERROR,
        'WARNING': logging.WARNING,
        'INFO': logging.INFO,
        'DEBUG': logging.DEBUG,
        'NOTSET': logging.NOTSET,
    }

    loglevel = loglevels[loglevel]

    if logfile is not None:  # log to file
        logging.basicConfig(level=loglevel, datefmt=date_format,
                            format=log_format, filename=logfile)
    elif loglevel is not None:  # log to stdout
        logging.basicConfig(level=loglevel, datefmt=date_format,
                            format=log_format, stream=sys.stdout)
        LOGGER.debug('Logging initialized')


def urlopen_(url: str):
    """
    Helper function for downloading a URL

    :param url: URL to download

    :returns: `http.client.HTTPResponse`
    """

    try:
        response = urlopen(url)
    except (ssl.SSLError, URLError) as err:
        LOGGER.warning(err)
        LOGGER.warning('Creating unverified context')
        context = ssl._create_unverified_context()

        response = urlopen(url, context=context)

    return response



###############################################################################
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
###############################################################################

import csv
import logging
from pathlib import Path
import re
from typing import List

import click

from pywis_topics.util import get_cli_common_options, get_userdir, setup_logger

LOGGER = logging.getLogger(__name__)


WIS2_TOPIC_HIERARCHY_LOOKUP = Path(get_userdir()) / 'wis2-topic-hierarchy'


class TopicHierarchy:
    def __init__(self, tables: str = None):
        """
        Initializer

        :param tables: location of base directory for bundle

        :returns: `pywis_topics.topics_TopicHierarchy`
        """

        self.topics = []

        topic_levels = [
            'channel',
            'version',
            'system',
            'centre-id',
            'notification-type',
            'data-policy',
            'earth-system-discipline'
        ]

        if tables is not None:
            tables_dir = Path(tables) / 'wis2-topic-hierarchy'
        else:
            tables_dir = WIS2_TOPIC_HIERARCHY_LOOKUP

        for topic_level in topic_levels:
            filename = tables_dir / f'{topic_level}.csv'
            with filename.open() as fh:
                level_topics = []
                reader = csv.reader(fh)
                next(reader)
                for row in reader:
                    level_topics.append(row[0])

                self.topics.append(level_topics)

    def list_children(self, topic_hierarchy: str = None) -> List[str]:
        """
        Lists children at a given level of a topic hierarchy

        :param topic_hierarchy: `str` of topic hierarchy

        :returns: `list` of topic children
        """

        matches = []

        if topic_hierarchy == '/':
            LOGGER.debug('Dumping root topic children')
            return self.topics[0]

        if not self.validate(topic_hierarchy):
            msg = 'Invalid topic'
            LOGGER.info(msg)
            raise ValueError(msg)

        th_tokens = topic_hierarchy.split('/', 6)
        num_th_tokens = len(th_tokens)

        if num_th_tokens < 6:
            LOGGER.debug('Listing core topics')
            subtopics = set(self.topics[len(th_tokens)])
        elif num_th_tokens == 6:
            LOGGER.debug('Listing earth system discipline topics')
            subtopics = set([t.split('/')[0] for t in self.topics[6]])
        else:
            LOGGER.debug('Listing domain topics')
            subtopics_to_add = []
            domain_topic = th_tokens[-1]
            for subtopic in self.topics[-1]:
                if subtopic.startswith(domain_topic):
                    if subtopic != domain_topic:
                        mask = subtopic.replace(f'{domain_topic}/', '')
                        if mask:
                            subtopics_to_add.append(mask.split('/')[0])

            subtopics = set(subtopics_to_add)

        matches.extend(subtopics)

        if not matches:
            msg = f'No matching topics for {topic_hierarchy}'
            LOGGER.info(msg)
            raise ValueError(msg)

        return matches

    def validate(self, topic_hierarchy: str = None,
                 strict: bool = True) -> bool:
        """
        Validates a topic hierarchy

        :param topic_hierarchy: `str` of topic hierarchy
        :param strict: `bool` of whether to perform strict validation,
                       including centre-id

        :returns: `bool` of whether topic hierarchy is valid
        """

        LOGGER.debug(f'Validating topic hierarchy {topic_hierarchy}')

        if topic_hierarchy in ['/', None]:
            msg = 'Topic hierarchy is empty'
            LOGGER.warning(msg)
            raise ValueError(msg)

        validate_baseline(topic_hierarchy)

        all_tokens = topic_hierarchy.split('/')
        core_tokens = all_tokens[:6]
        esd_subtopic = '/'.join(all_tokens[6:])

        LOGGER.debug(f'Core tokens: {core_tokens}')
        LOGGER.debug(f'Earth system discipline subtopic: {esd_subtopic}')
        LOGGER.debug('Validating core tokens')

        if not self._validate_core(core_tokens, strict):
            LOGGER.debug('Core tokens are invalid')
            return False

        if esd_subtopic:
            LOGGER.debug('Validating Earth system discipline subtopic')
            if not self._validate_esd_subtopic(esd_subtopic, strict):
                LOGGER.debug('Earth system discipline subtopic is invalid')
                return False
        else:
            LOGGER.debug('No Earth system discipline subtopic')

        return True

    def _validate_core(self, core_tokens: list, strict: bool = True) -> bool:
        """
        Validates core topic tokens

        :param core_tokens: `list` of core tokens
        :param strict: `bool` of whether to perform strict validation,
                       including centre-id

        :returns: `bool` of whether topic hierarchy is valid
        """

        for count, value in enumerate(core_tokens):
            if value in [None, '']:
                continue
            if not strict and count == 3:
                LOGGER.debug('Skipping centre-id validation')
                continue
            elif value in ['+', '#']:
                if not strict:
                    LOGGER.debug('Skipping wildcard')
                    continue
                else:
                    return False

            if value not in self.topics[count]:
                return False

        return True

    def _validate_esd_subtopic(self, esd_subtopic: str,
                               strict: bool = True) -> bool:
        """
        Validates Earth system discipline subtopic

        :param esd_subtopic: `str` of Earth system discipline subtopic
        :param strict: `bool` of whether to perform strict validation

        :returns: `bool` of whether topic hierarchy is valid
        """

        is_valid = False

        if strict:
            LOGGER.debug('Validating subtopic with strict mode')
            is_experimental = esd_subtopic.split('/')[1] == 'experimental'
            return esd_subtopic in self.topics[-1] or is_experimental

        tokens = esd_subtopic.split('/')
        if len(tokens) > 1 and tokens[1] == 'experimental':
            LOGGER.debug('Experimental topic found, skipping')
            return True

        regex = esd_subtopic
        if '+' in regex:
            regex = regex.replace('+', r'(\w+|\+)')
        if '#' in regex:
            regex = regex.replace('#', r'.*')

        regex = f'^{regex}$'

        LOGGER.debug(regex)

        for esd in self.topics[-1]:
            LOGGER.debug(f'Testing {esd} against {regex}')
            match = re.search(regex, esd)
            if not match:
                LOGGER.debug('No match')
            else:
                LOGGER.debug('Match')
                last_token_match = False
                if not esd.endswith('#'):
                    esd_last_token = esd.split('/')[-1]
                    regex_last_token = (regex.split('/')[-1].replace('$', '')
                                        .replace('.*', '')
                                        .replace('^', ''))
                    if regex_last_token == '':
                        last_token_match = True
                    elif esd_last_token == regex_last_token:
                        last_token_match = True

                if last_token_match:
                    is_valid = True

        return is_valid


def validate_baseline(topic_hierarchy: str = None) -> bool:
    """
    Validates a topic hierarchy baseline conventions

    :param topic_hierarchy: `str` of topic hierarchy

    :returns: `bool` of whether topic hierarchy is valid to the baseline
              conventions
    """

    if '.' in topic_hierarchy:
        msg = 'Topic cannot contain dots'
        LOGGER.warning(msg)
        return False

    if not topic_hierarchy.islower():
        msg = 'Topic must be lowercase'
        LOGGER.warning(msg)
        return False

    if not topic_hierarchy.isascii():
        msg = 'Topic must be IRA T.50'
        LOGGER.warning(msg)
        return False

    if '#' in topic_hierarchy:
        pos = topic_hierarchy.find('#')
        if pos != len(topic_hierarchy) - 1:
            LOGGER.warning('Multi-level wildcard can only be last')
            return False

    return True


@click.group()
def topic():
    """Topic hierarchy utilities"""
    pass


@click.command('list')
@click.pass_context
@get_cli_common_options
@click.argument('topic-hierarchy')
def list_(ctx, topic_hierarchy, logfile, verbosity):
    """List topic hierarchies at a given level"""

    setup_logger(verbosity, logfile)

    th = TopicHierarchy()

    try:
        matching_topics = th.list_children(topic_hierarchy)
        click.echo('Matching topics')
        for matching_topic in matching_topics:
            click.echo(f'- {matching_topic}')
    except ValueError as err:
        raise click.ClickException(err)


@click.command()
@click.pass_context
@get_cli_common_options
@click.option('--strict/--no-strict', default=True,
              help='Validate in strict mode')
@click.argument('topic-hierarchy')
def validate(ctx, topic_hierarchy, logfile, verbosity, strict=True):
    """Validate topic hierarchy"""

    setup_logger(verbosity, logfile)

    th = TopicHierarchy()

    if th.validate(topic_hierarchy, strict=strict):
        click.echo('Valid')
    else:
        click.echo('Invalid')


topic.add_command(list_)
topic.add_command(validate)
